#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <stdio.h>

#include "lcd.h"
#include "Keypad.h"
#include "millis.h"

#define pinIN(port, ddr, pin) \
		(ddr) &= ~(1 << (pin));	\
		(port) |= (1<<(pin));
		
#define pinOUT(port, ddr, pin) \
		(ddr) |= (1<<(pin));	
		
#define set_on(port, pin)\
		(port) |= (1 << (pin));
		
#define set_off(port, pin)\
		(port) &= ~(1 << (pin));

#define BUTTON_PRESSED(pinblock ,pin)	\
		(pinblock & (1 << pin))

#define READINGS 200
#define GOOD_READINGS 170

// Define time to activate alarm in ms
#define TIME_TO_ACTIVATE 2000

// Define time until detection and call the cops ^^
#define TIME_TO_DETECT 5000

#define ALARM_CODE 1471
#define ALARM_CODE_LENGTH 4

#define ALARM_PIN PB1
#define ALARM_PORT PORTB
#define ALARM_DDR DDRB
#define ALARM_PINBLOCK PINB

#define PIR_PIN PD0
#define PIR_PORT PORTD
#define PIR_DDR DDRD
#define PIR_PINBLOCK PIND

enum STATE { INIT, CODE_PROMPT, IDLE, OPTIONS, ALARM_ON, ALARM_OFF } State, lastState;

uint8_t readings, motionStarted, motionDetected;
millis_t motionStartedTime, alarmTime;
Keypad keypad;
char key;


void activatePIR()
{
	PCICR |= (1 << PCIE3);
	PCMSK3 |= (1 << PCINT24);
}

void deactivatePIR()
{
	PCICR &= ~(1 << PCIE3);
	PCMSK3 &= ~(1 << PCINT24);
	set_off(PORTD, PD7);
}

void activateAlarm()
{
	activatePIR();
	lastState = State;
	State = ALARM_ON;
}

void deactivateAlarm()
{
	deactivatePIR();
	lastState = State;
	State = ALARM_OFF;
}

void turnAlarmOn()
{
	set_off(ALARM_PORT, ALARM_PIN);
}

void turnAlarmOff()
{
	set_on(ALARM_PORT, ALARM_PIN);
}

ISR(PCINT3_vect)
{
	// Read PIR
	cli();
	_delay_ms(1500);
	if( BUTTON_PRESSED(PIR_PINBLOCK, PIR_PIN) )
	{
		readings = 0;
		for (int i = 0; i < READINGS; i++) {
			if( BUTTON_PRESSED(PIR_PINBLOCK, PIR_PIN) )
				readings ++;
			_delay_ms(1);
		}

		if(readings > GOOD_READINGS) {
			set_on(PORTD, PD7);
			motionStarted = 1;
			motionStartedTime = millis();
		}
	}
	else {
		set_off(PORTD, PD7);
		motionStarted = 0;
	}
	sei();
}


uint32_t promptForCode()
{
	LCD_clear();
	LCD_print("Enter code:");
	uint32_t code;
	keypad.resetAll();
	
	if(State != CODE_PROMPT) {
		lastState = State;
		State = CODE_PROMPT;
	}
	
	while(1) {		
		if (keypad.GetKeyPressed() == KEY_ESC && lastState == IDLE) {
			State = IDLE;
			lastState = CODE_PROMPT;
			return 0;
		} 
		keypad.readCode();
		code = keypad.getCode();
		if(keypad.canReturn()) {
			if(code > 999)
				return code;
			else {
				LCD_clear();
				LCD_print("Code too short!");
				_delay_ms(1500);
			}
		}
	}
}


void computeAlarm(millis_t now)
{
	if(motionDetected || key == KEY_STAR) {
		uint32_t code = promptForCode();

		if( code == ALARM_CODE) {					
			motionDetected = 0;
			// deactivates motion detection
			deactivateAlarm();
			_delay_ms(100);
			// turns off the siren
			turnAlarmOff();
			_delay_ms(100);
		}
		else if (code != 0) {
			LCD_clear();

			LCD_print("   Try again");
			if(key == KEY_STAR)
				LCD_printAt(0x40, " Press * again");
			
			lastState = State;
			State = ALARM_ON;
						
			_delay_ms(1500);
			return;
		}
		else {
			lastState = State;
			State = IDLE;
		}
	
	}
	
	if(now - motionStartedTime > TIME_TO_DETECT && motionStarted)
	{
		motionStarted = 0;
		turnAlarmOn();
		
		motionDetected = 1;
		alarmTime = now;
	}
}

void displayOptions()
{
	if(State == OPTIONS && lastState != OPTIONS)
	{
		LCD_clear();
		LCD_print("  Options menu");
		LCD_printAt(0x40, "  Esc = Return");
		lastState = State;
	}
}
void displayIdle()
{
	if(State == IDLE && lastState != IDLE) {
		LCD_clear();
		LCD_print("  F1 = Options");
		LCD_printAt(0x40, " ENT = Activate");
		lastState = State;
	}
}
void handleOptionsMode(char &key)
{
	switch(key) {
		case KEY_ESC:
			lastState = State;
			State = IDLE;
			break;
		default:
			displayOptions();
	}
}
void handleIdleMode(char &key)
{
	uint32_t code;
	//LCD_clear();
	
	switch(key) {
		case KEY_ENTER:
			LCD_clear();
			_delay_ms(10);
			code = promptForCode();
			
			if(code == ALARM_CODE) {
				LCD_clear();
				LCD_print("Alarm activates");
				LCD_printAt(0x40, "in 2 seconds");
				_delay_ms(TIME_TO_ACTIVATE);
				
				activateAlarm();
				
				LCD_clear();
				LCD_print("Alarm activated!");
				LCD_printAt(0x40, " *  deactivates");
			}
			else if (code != 0) {
				LCD_clear();
				LCD_print("     Wrong code");
				LCD_printAt(0x40, "Press ENT again");
		
				lastState = State;
				State = IDLE;
				
				_delay_ms(1500);
			}
			else { 
				lastState = CODE_PROMPT;
				State = IDLE;
			}
				
			break;
		case KEY_F1:
			lastState = State;
			State = OPTIONS;
			break;
		default:
			displayIdle();
			break;
	}
}

int main()
{	

	lastState = INIT;
	
	//Wait for LCD To Start
	_delay_loop_2(10);

	//Now initialize the modules
	
	LCD_init();
	millis_init();

	// Set user LED out
	pinOUT(PORTD, DDRD, PD7);
	
	// Set alarm relay OUT
	pinOUT(ALARM_PORT, ALARM_DDR, ALARM_PIN);
	
	// Set alarm OFF
	turnAlarmOff();
	
	// Set port for PIR sensor	
	pinIN(PIR_PORT, PIR_DDR, PIR_PIN);

	// Activate interrupts
	sei();
	
	// Be sure that alarm is off
	deactivateAlarm();
	
	// Wait for modules setup
	_delay_ms(1000);
	
	LCD_print("    Welcome!");
	LCD_printAt(0x40, "Don't move quick");
	
	_delay_ms(1500);
	
	State = IDLE;
	
	while(1) {
		key=keypad.GetKeyPressed(); //Get the keycode of pressed key
		
		millis_t now = millis();
		switch (State) {
			case INIT:
				break;
			case IDLE:
				handleIdleMode(key);
				break;
			case ALARM_ON:
				computeAlarm(now);
				break;
			case ALARM_OFF:				
				lastState = State;
				State = IDLE;
				
				LCD_clear();
				LCD_print("Alarm stopped");
				_delay_ms(2000);
				break;
			case OPTIONS:
				handleOptionsMode(key);
				break;
		}	
	}
	return 0;
}
